/*
** Copyright (C) 1999-2001 Christian Werner
**
** This program and library is free software; you can redistribute it and/or
** modify it under the terms of the GNU (Library) General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
** GNU (Library) General Public License for more details.
**
** You should have received a copy of the GNU (Library) General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package swishej;

import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.*;
import java.applet.*;
import netscape.javascript.JSObject;

/**
 * Class implementing the search applet for SWISH-E indices.
 */

public class SearchApplet extends Applet {
    /**
     * The "Search" button.
     */
    Button doSearch;

    /**
     * The "Words" button.
     */
    Button doWords;

    /**
     * The "Clear" button.
     */
    Button doClear;

    /**
     * The "Help" button.
     */
    Button doHelp;

    /**
     * The keywords label field.
     */
    Label label;

    /**
     * The keyword entry field.
     */
    TextField words;

    /**
     * Panel for the user interface elements.
     */
    Panel p;

    /**
     * URL of the SWISH-E index file.
     */
    URL indexname;

    /**
     * Loaded index file representation.
     */
    IndexFile inf;

    /**
     * JavaScript window (= frame) object for displaying the search result.
     */
    JSObject win;

    /**
     * String array for passing arguments to JavaScript.
     */
    String data[];

    /**
     * Character mapping.
     */
    char map[] = null;

    /**
     * Applet initialization.
     * <ol>
     * <li> Applet parameters are retrieved.
     * <li> The JavaScript window object for displaying the result is
     *      obtained and verified.
     * <li> The URLs for the index file and help page are retrieved.
     * <li> The user interface is built (3 buttons, one text entry)
     * <li> Some JavaScript variables and functions are defined. These
     *      are needed later to display search results, to clear the
     *      result frame, and to display the help page.
     * </ol>
     */
    public void init() {
        String resultframe = getParameter("resultframe");
	URL helpname = null;
	indexname = null;
	inf = null;
	win = null;
	data = new String[1];

	p = new Panel();
	p.setLayout(new FlowLayout(FlowLayout.CENTER, 4, 4));

	label = new Label();
	label.setAlignment(Label.RIGHT);

	try {
	    win = JSObject.getWindow(this);
	} catch (Exception e) {
	    System.out.println("can't get JS window of applet:\n" + e);
	    label.setText("JavaScript Error, can't continue !");
	    p.add(label);
	    add(p);
	    validate();
	    return;
	}
	try {
	    win.eval("rwin = parent." + resultframe);
	    win.eval("ismozilla = (navigator.appName == \"Netscape\")");
	    win.eval("if (ismozilla) {\n" +
		     "    rdoc = rwin.document;\n" +
		     "} else {\n" +
		     "    dummy = rwin.document;\n" +
		     "}");
	} catch (Exception e) {
	    label.setText("JavaScript Error, can't find resultframe !");
	    p.add(label);
	    add(p);
	    validate();
	    win = null;
	    return;
	}
        try {
	    indexname = new URL(getDocumentBase(),
				getParameter("indexfile"));
	} catch (MalformedURLException e) {
	    System.out.println("bad URL for indexfile:\n" + e);
	}
	try {
	    helpname = new URL(getDocumentBase(),
			       getParameter("helpfile"));
	} catch (MalformedURLException e) {
	    System.out.println("bad URL for helpfile:\n" + e);
	}

	String s;
	s = getParameter("kwlabel");
	if (s == null) {
	    label.setText("Keywords:");
	} else if (s.length() == 0) {
	    label = null;
	} else {
	    label.setText(s);
	}   
	words = new TextField("", 35);
	doSearch = new Button();
	s = getParameter("searchbutton");
	if (s == null) {
	    doSearch.setLabel("Search");
	} else if (s.length() == 0) {
	    doSearch = null;
	} else {
	    doSearch.setLabel(s);
	}
	doWords = new Button();
	s = getParameter("wordsbutton");
	if (s == null) {
	    doWords.setLabel("All Words");
	} else if (s.length() == 0) {
	    doWords = null;
	} else {
	    doWords.setLabel(s);
	}
	doClear = new Button();
	s = getParameter("clearbutton");
	if (s == null) {
	    doClear.setLabel("Clear");
	} else if (s.length() == 0) {
	    doClear = null;
	} else {
	    doClear.setLabel(s);
	}
	doHelp = new Button();
	s = getParameter("helpbutton");
	if (s == null) {
	    doHelp.setLabel("Help");
	} else if (s.length() == 0) {
	    doHelp = null;
	} else {
	    doHelp.setLabel(s);
	}
	s = getParameter("charmap");
	if (s != null) {
	    map = new char[256];
	    int i, k, n, x;
	    for (i = 0; i < 256; i++) {
		map[i] = (char) i;
	    }
	    for (i = k = n = x = 0; i < 256;) {
		char c = s.charAt(k);
		if (c >= '0' && c <= '9') {
		    x = x << 4;
		    x |= c - '0';
		    n++;
		} else if (c >= 'A' && c <= 'F') {
		    x = x << 4;
		    x |= c - 'A' + 10;
		    n++;
		} else if (c >= 'a' && c <= 'f') {
		    x = x << 4;
		    x |= c - 'a' + 10;
		    n++;
		}
		if (n == 4) {
		    map[i] = (char) x;
		    i++;
		    n = 0;
		}
		++k;
		if (k >= s.length()) {
		    break;
		}
	    }
	}

	win.eval("function putstr(txt) {\n" +
		 "    if (ismozilla) {\n" +
		 "        rdoc.close();\n" +
		 "    } else {\n" +
		 "        rdoc = rwin.document;\n" +
		 "        rdoc.open();\n" +
		 "    }\n" +
		 "    rdoc.write(txt);\n" +
		 "    rdoc.close();\n" +
		 "}");
	win.eval("function clrdoc(txt) {\n" +
		 "    if (ismozilla) {\n" +
		 "        rdoc.close();\n" +
		 "    } else {\n" +
		 "        rdoc = rwin.document;\n" +
		 "        rdoc.open();\n" +
		 "    }\n" +
		 "    rdoc.write(\"<body></body>\");\n" +
		 "    rdoc.close();\n" +
		 "}");
	win.eval("function dohelp() {\n" +
		 "    rwin.location = \"" + helpname.toString() + "\";\n" +
		 "}");
    }

    /**
     * Applet startup. The index file is read.
     */
    public void start() {
        if (win == null) {
	    return;
	}
	if (label != null) {
	    p.add(label);
	}
	p.add(words);
	if (doSearch != null) {
	    p.add(doSearch);
	}
	if (doWords != null) {
	    p.add(doWords);
	}
	if (doClear != null) {
	    p.add(doClear);
	}
	if (doHelp != null) {
	    p.add(doHelp);
	}
	add(p);
	validate();
        if (inf == null) {
	    showStatus("Loading index ...");
	    try {
	        inf = Search.loadindex(indexname);
	    } catch (Exception e) {
	        showStatus("Error loading index:\n" + e);
	    }
	    showStatus("Index " + (inf == null ? "not" : "") + "loaded.");
	} else {
	    showStatus("Ready for search.");
	}
    }

    /**
     * Applet stop.
     */
    public void stop() {
	invalidate();
	p.removeAll();
	removeAll();
    }

    /**
     * Applet destruction. The index file reference is set to null.
     */
    public void destroy() {
        inf = null;
	win = null;
    }

    /**
     * Applet action method for starting a keyword search, clearing
     * the result frame, or displaying the help page in the result frame.
     */
    public boolean action(Event ev, Object obj) {
        if ((ev.target == doSearch || ev.target == words) && inf != null) {
	    showStatus("Searching ...");
	    String result = Search.search(words.getText(), -1, -1, inf,
					  getDocumentBase());
	    data[0] = remap(map, result);
	    win.call("putstr", data);
	    showStatus("Done.");
	} else if (ev.target == doWords) {
	    words.setText("");
	    showStatus("Making word list ...");
	    String result = fmtallwords();
	    data[0] = result;
	    win.call("putstr", data);
	    showStatus("Done.");
	} else if (ev.target == doClear) {
	    words.setText("");
	    data[0] = null;
	    win.call("clrdoc", data);
	    showStatus("");
	} else if (ev.target == doHelp) {
	    data[0] = null;
	    win.call("dohelp", data);
	    showStatus("");
	}
	return true;
    }

    /**
     * Add word to word text input field
     * @param w word to be added
     */
    public void addword(String w) {
        if (win != null && words != null) {
	    words.setText(words.getText() + " " + w);
	}
    }

    /**
     * Clear text input field.
     */
    public void clrwords() {
        if (win != null && words != null) {
	    words.setText("");
	}
    }

    /**
     * Display word list given start character
     * @param w string whose first character is taken as start character
     */
    public void showwords(String w) {
        if (win != null) {
	    showStatus("Making word list ...");
	    String result = fmtwords(w);
	    data[0] = result;
	    win.call("putstr", data);
	    showStatus("Done.");
	}
    }

    /**
     * Request additional space for 3D border.
     * @return Insets with 2 pixel border.
     */
    public Insets insets() {
	return new Insets(2, 2, 2 + 1, 2 + 1);
    }

    /**
     * Paint method of applet panel, draws 3D border.
     * @param g Graphics object
     */
    public void paint(Graphics g) {
	Dimension d = size();
	Color bg = getBackground();
	super.paint(g);
	g.setColor(bg); 
	g.draw3DRect(0, 0, d.width - 1, d.height - 1, true);
    }

    /**
     * Format keywords as HTML
     * @return HTML formatted word list
     */ 
    String fmtallwords() {
        return fmtwords(IndexFile.indexchars);
    }

    /**
     * Format keywords as HTML
     * @param String which first character determines sublist output
     * @return HTML formatted word list
     */ 
    String fmtwords(String first) {
	if (first.length() > 1) {
	    if (first.charAt(0) == '#') {
		try {
		    Integer xi =
			new Integer(Integer.parseInt(first.substring(1)));
		    char cb[] = new char[1];
		    cb[0] = (char) xi.intValue();
		    first = new String(cb);
		} catch (Exception e) {
		}
	    }
	}
        String words[] = inf.dumpWords(first);
	String indexchars = IndexFile.indexchars;
	StringBuffer sb = new StringBuffer();
	sb.append("<base href=\"\">\n");
	sb.append("<script language=\"javascript\">\n");
	/*
	 * Transfer name of applet window.
	 */ 
	sb.append("awin = parent." + (String) win.getMember("name") + ";\n");
	/*
	 * Find the search applet and transfer its number assuming
	 * that the JavaScript engine counts as getApplets() enumerates.
	 */
	Enumeration e = getAppletContext().getApplets();
	int count = 0;
	boolean found = false;
	while (e.hasMoreElements()) {
	    if (this.equals((Applet) e.nextElement())) {
	        sb.append("anum = " + count + ";\n");
		found = true;
		break;
	    }
	}
	if (!found) {
	    /*
	     * Fallback: use first applet.
	     */
	    sb.append("anum = 0;\n");
	    System.out.println("Cannot find applet");
	}
	/*
	 * Define JavaScript functions for
	 *  - adding a keyword into the applet's text input field
	 *  - reentering this method for another initial
	 *  - clearing the applet's text input field
	 */
	sb.append("function x(w) {\n" +
		  "    w = unescape(w);\n" +
		  "    awin.document.applets[anum].addword(w);\n" +
		  "}\n");
	sb.append("function g(w) {\n" +
		  "    awin.document.applets[anum].showwords(w);\n" +
		  "}\n");
	sb.append("function c() {\n" +
		  "    awin.document.applets[anum].clrwords();\n" +
		  "}\n");
	sb.append("\n</script><body>\n");
	sb.append("<a href='javascript:x(\"AND\")'>AND</a>\n");
	sb.append("<a href='javascript:x(\"OR\")'>OR</a>\n");
	sb.append("<a href='javascript:x(\"NOT\")'>NOT</a>\n");
	sb.append("<a href='javascript:c()'>CLEAR</a><hr>\n");
	char firstc = first.charAt(0);
	int x = firstc;
	sb.append("<h1>&#" + remap(map, x) + ";</h1>\n");
	for (int k = 0; k < indexchars.length(); k++) {
	    x = indexchars.charAt(k);
	    if (x == firstc) {
	        continue;
	    }
	    sb.append("<a href='javascript:g(\"#" + (x & 0xFF) + "\")'>");
	    sb.append("&#" + remap(map, x) + ";</a>\n");
	}
	sb.append("<br><hr>");
        count = 0;
	for (int i = 0; words != null && i < words.length; i++) {
	    String w = words[i];
	    sb.append("<a href=\"javascript:x('" +
		      URLEncoder.encode(w) + "')\">");
	    w = remap(map, w);
	    for (int k = 0; k < w.length(); k++) {
	        x = w.charAt(k);
		if (x < 32 || x > 127) {
		    sb.append("&#" + x + ";");
		} else {
		    sb.append((char) x);
		}
	    }
	    sb.append("</a> ");
	    if (++count > 5) {
	        sb.append("<br>");
		count = 0;
	    }
	}
	sb.append("</body>");
  	return sb.toString();
    }

    /**
     * Return information about applet parameters.
     * @return String array with parameter description.
     */
    public String[][] getParameterInfo() {
        String[][] info = {
	    {"resultframe", "string", "browser frame for result"},
	    {"indexfile", "url", "swish-e index file"},
	    {"helpfile", "url", "file to display help"},
	    {"searchbutton", "string", "text for search button"},
	    {"wordsbutton", "string", "text for words button"},
	    {"clearbutton", "string", "text for clear button"},
	    {"helpbutton", "string", "text for help button"},
	    {"kwlabel", "string", "text for keyword entry label"},
	    {"charmap", "string", "character mapping for result pages"}
	};
	return info;
    }

    /**
     * Applet info blurks.
     * @return string with copy.*t information.
     */
    public String getAppletInfo() {
        return "SWISH-E Search Applet\n" +
	    "Copyright (C) 1999-2001 Christian Werner\n\n" +
	    "This program and library is free software; you can\n" +
	    "redistribute it and/or modify it under the terms of\n" +
	    "the GNU (Library) General Public License as published\n" +
	    "by the Free Software Foundation; either version 2 of\n" +
	    "the License, or any later version.\n\n" +
	    "This program is distributed in the hope that it will be\n" +
	    "useful, but WITHOUT ANY WARRANTY; without even the\n" +
	    "implied warranty of MERCHANTABILITY or FITNESS FOR A\n" +
	    "PARTICULAR PURPOSE. See the GNU (Library) General Public\n" +
	    "License for more details.\n\n" +
	    "You should have received a copy of the GNU (Library)\n" +
	    "General Public License along with this program; if not,\n" +
	    "write to the Free Software Foundation, Inc., 59 Temple\n" +
	    "Place - Suite 330, Boston, MA  02111-1307, USA.\n";
    }

    /**
     * Character mapping function.
     * @param i integer to be mapped
     * @return mapped integer
     */
    int remap(char map[], int i) {
	if (map == null) {
	    return i & 0xFF;
	}
	return (int) map[i];
    }

    /**
     * Character mapping function.
     * @param s String to be mapped
     * @return mapped string
     */
    String remap(char map[], String s) {
	StringBuffer sb = new StringBuffer();
	for (int i = 0; i < s.length(); i++) {
	    char c = s.charAt(i);
	    if (map == null) {
		c = (char) ((int) c & 0xFF);
	    } else {
		c = map[(int) c];
	    }
	    sb.append(c);
	}
	return sb.toString();
    }

    /**
     * If started as application, this method runs a search on
     * the local file "index.swish-e" with the keywords taken from
     * the command line.
     * @param args command line
     */
    public static void main(String[] args) {
        StringBuffer words = new StringBuffer();
	String sep = "";
	for (int i = 0; i < args.length; i++) {
	    words.append(sep);
	    words.append(args[i]);
	    sep = " ";
	}
	System.out.println(Search.search(words.toString(), -1, -1,
					 "index.swish-e"));
	System.exit(0);
    }
}

